import unittest
from unittest.mock import patch, Mock
from requests import Session

from crodl.tools.scrap import (
    AudioUUIDDoesNotExist,
    PageDoesNotExist,
    DataEntryDoesNotExist,
    PlayerWrapperDoesNotExist,
    ShowUUIDDoesNotExist,
    get_attributes,
    get_audio_link_of_preferred_format,
    get_audio_uuid,
    get_js_value_from_url,
    get_series_id,
    get_show_uuid,
    is_series,
    is_show,
)


class TestGetAudioUUID(unittest.TestCase):
    @patch.object(Session, "get")
    def test_successful_retrieval(self, mock_get):
        site_url = "https://example.com"
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = (
            '<section class="player-wrapper" data-entry=\'{"uuid": "12345"}\'>'
        )
        mock_get.return_value = mock_response

        uuid = get_audio_uuid(site_url, Session())
        self.assertEqual(uuid, "12345")

    @patch.object(Session, "get")
    def test_handle_404(self, mock_get):
        site_url = "https://example.com"
        mock_response = Mock()
        mock_response.status_code = 404
        mock_get.return_value = mock_response

        with self.assertRaises(PageDoesNotExist):
            get_audio_uuid(site_url, Session())

    @patch.object(Session, "get")
    def test_handle_player_wrapper_not_found(self, mock_get):
        site_url = "https://example.com"
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = "<div></div>"
        mock_get.return_value = mock_response

        with self.assertRaises(PlayerWrapperDoesNotExist):
            get_audio_uuid(site_url, Session())

    @patch.object(Session, "get")
    def test_handle_uuid_not_found(self, mock_get):
        site_url = "https://example.com"
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = (
            '<section class="player-wrapper" data-entry=\'{"invalid_key": "12345"}\'>'
        )
        mock_get.return_value = mock_response

        with self.assertRaises(AudioUUIDDoesNotExist):
            get_audio_uuid(site_url, Session())

    @patch.object(Session, "get")
    def test_data_entry_not_found(self, mock_get):
        site_url = "https://example.com"
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = '<section class="player-wrapper">'
        mock_get.return_value = mock_response
        with self.assertRaises(DataEntryDoesNotExist):
            get_audio_uuid(site_url, Session())

    @patch.object(Session, "get")
    def test_data_entry_is_empty_string(self, mock_get):
        site_url = "https://example.com"
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = '<section class="player-wrapper" data-entry="">'
        mock_get.return_value = mock_response
        with self.assertRaises(DataEntryDoesNotExist):
            get_audio_uuid(site_url, Session())


class TestGetShowUUID(unittest.TestCase):
    @patch.object(Session, "get")
    def test_successful_retrieval(self, mock_get):
        site_url = "https://example.com"
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = (
            '<div class="b-detail" data-entry=\'{"show-uuid": "12345"}\'></div>'
        )
        mock_get.return_value = mock_response

        uuid = get_show_uuid(site_url, Session())
        self.assertEqual(uuid, "12345")

    @patch.object(Session, "get")
    def test_handle_404(self, mock_get):
        site_url = "https://example.com"
        mock_response = Mock()
        mock_response.status_code = 404
        mock_get.return_value = mock_response

        with self.assertRaises(PageDoesNotExist):
            get_show_uuid(site_url, Session())

    @patch.object(Session, "get")
    def test_handle_uuid_not_found(self, mock_get):
        site_url = "https://example.com"
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = (
            '<div class="b-detail" data-entry=\'{"invalid_key": "12345"}\'></div>'
        )
        mock_get.return_value = mock_response

        with self.assertRaises(ShowUUIDDoesNotExist):
            get_show_uuid(site_url, Session())

    @patch.object(Session, "get")
    def test_data_entry_not_found(self, mock_get):
        site_url = "https://example.com"
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = '<div class="b-detail"></div>'
        mock_get.return_value = mock_response
        with self.assertRaises(DataEntryDoesNotExist):
            get_show_uuid(site_url, Session())

    @patch.object(Session, "get")
    def test_data_entry_is_empty_string(self, mock_get):
        site_url = "https://example.com"
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.text = '<div class="b-detail" data-entry=""></div>'
        mock_get.return_value = mock_response
        with self.assertRaises(DataEntryDoesNotExist):
            get_show_uuid(site_url, Session())


class TestGetAttributes(unittest.TestCase):
    @patch.object(Session, "get")
    def test_successful_retrieval(self, mock_get):
        mock_response = Mock()
        mock_response.json.return_value = {"data": {"attributes": "test_attributes"}}
        mock_get.return_value = mock_response

        attributes = get_attributes("12345", Session())
        self.assertEqual(attributes, "test_attributes")

    @patch.object(Session, "get")
    def test_handle_response_without_data(self, mock_get):
        mock_response = Mock()
        mock_response.json.return_value = {}
        mock_get.return_value = mock_response

        with self.assertRaises(AttributeError):
            get_attributes("12345", Session())

    @patch.object(Session, "get")
    def test_handle_missing_attributes_key(self, mock_get):
        mock_response = Mock()
        mock_response.json.return_value = {"data": {}}
        mock_get.return_value = mock_response

        result = get_attributes("12345", Session())
        self.assertEqual(result, {})


class TestGetAudioLinkOfPreferredFormat(unittest.TestCase):
    @patch("crodl.tools.scrap.get_preferred_audio_format")
    def test_preferred_format_found(self, mock_get_preferred_audio_format):
        attrs = {
            "audioLinks": [
                {"variant": "mp3", "url": "audio1.mp3"},
                {"variant": "aac", "url": "audio2.aac"},
            ]
        }
        mock_format = mock_get_preferred_audio_format.return_value
        mock_format.value = "aac"
        result = get_audio_link_of_preferred_format(attrs)
        self.assertEqual(result, "audio1.mp3")

    @patch("crodl.tools.scrap.get_preferred_audio_format")
    def test_no_audio_links(self, mock_get_preferred_audio_format):
        attrs = {"audioLinks": []}
        mock_format = mock_get_preferred_audio_format.return_value
        mock_format.value = "aac"
        result = get_audio_link_of_preferred_format(attrs)
        self.assertEqual(result, None)

    @patch("crodl.tools.scrap.get_preferred_audio_format")
    def test_preferred_format_not_found(self, mock_get_preferred_audio_format):
        attrs = {
            "audioLinks": [
                {"variant": "wma", "url": "audio1.wma"},
                {"variant": "ogg", "url": "audio2.ogg"},
            ]
        }
        mock_format = mock_get_preferred_audio_format.return_value
        mock_format.value = "aac"
        result = get_audio_link_of_preferred_format(attrs)
        self.assertIsNone(result)


class TestGetJsValueFromUrl(unittest.TestCase):
    @patch.object(Session, "get")
    def test_get_js_value_from_url_success(self, mock_get):
        response = Mock()
        response.status_code = 200
        response.text = '<script>var dl = {"siteEntityBundle":"serial", "contentId":"1234"};</script>'
        mock_get.return_value = response

        result = get_js_value_from_url("http://example.com", "contentId", Session())
        self.assertEqual(result, "1234")

    @patch.object(Session, "get")
    def test_get_js_value_from_url_failure(self, mock_get):
        response = Mock()
        response.status_code = 404
        mock_get.return_value = response

        result = get_js_value_from_url("http://example.com", "jsvar", Session())
        self.assertIsNone(result)

    @patch.object(Session, "get")
    def test_get_js_value_from_url_failure_no_match(self, mock_get):
        response = Mock()
        response.status_code = 200
        response.text = '<script>var dl = {"contentId":};</script>'
        mock_get.return_value = response

        result = get_js_value_from_url("http://example.com", "conte", Session())
        self.assertIsNone(result)

    @patch.object(Session, "get")
    def test_get_js_value_from_url_failure_jsvar_not_in_text(self, mock_get):
        response = Mock()
        response.status_code = 200
        response.text = '<script>var dl = {"siteEntityBundle":"serial", "contentId":"1234"};</script>'
        mock_get.return_value = response

        result = get_js_value_from_url("http://example.com", "jsvar", Session())
        self.assertIsNone(result)


class TestIsSerial(unittest.TestCase):
    @patch("crodl.tools.scrap.get_js_value_from_url")
    def test_is_serial_true(self, mock_get_js_value_from_url):
        mock_get_js_value_from_url.return_value = "serial"
        result = is_series("http://example.com", Session())
        self.assertTrue(result)

    @patch("crodl.tools.scrap.get_js_value_from_url")
    def test_is_serial_false(self, mock_get_js_value_from_url):
        mock_get_js_value_from_url.return_value = "movie"
        result = is_series("http://example.com", Session())
        self.assertFalse(result)


class TestGetSeriesId(unittest.TestCase):
    @patch("crodl.tools.scrap.get_js_value_from_url")
    def test_valid_series_id(self, mock_get_js_value_from_url):
        mock_get_js_value_from_url.return_value = "31415"
        result = get_series_id("http://example.com", Session())
        self.assertEqual(result, "31415")


class TestIsShow(unittest.TestCase):
    @patch("crodl.tools.scrap.get_js_value_from_url")
    def test_is_show_true(self, mock_get_js_value_from_url):
        mock_get_js_value_from_url.return_value = "show"
        result = is_show("http://example.com", Session())
        self.assertTrue(result)

    @patch("crodl.tools.scrap.get_js_value_from_url")
    def test_is_show_false(self, mock_get_js_value_from_url):
        mock_get_js_value_from_url.return_value = "serial"
        result = is_show("http://example.com", Session())
        self.assertFalse(result)


if __name__ == "__main__":
    unittest.main()



import json
import re

from bs4 import BeautifulSoup, Tag
from requests import Session

from crodl.settings import API_SERVER, PREFERRED_AUDIO_FORMAT
from crodl.tools.logger import crologger
from crodl.streams.utils import get_preferred_audio_format
from crodl.exceptions import (
    PageDoesNotExist,
    PlayerWrapperDoesNotExist,
    DataEntryDoesNotExist,
    AudioUUIDDoesNotExist,
    ShowUUIDDoesNotExist,
)

cro_session = Session()


def get_audio_uuid(site_url: str, session: Session) -> str:
    """Return the audio UUID from the site URL."""

    crologger.info("Opening URL: %s", site_url)
    response = session.get(site_url, timeout=5)

    if response.status_code == 404:
        crologger.error("The page %s does not exist.", site_url)
        raise PageDoesNotExist("The page %s does not exist.")

    soup = BeautifulSoup(response.text, "html.parser")

    # UUID and other attributes are in the player-wrapper
    player_wrapper = soup.select(".player-wrapper")

    if not player_wrapper:
        raise PlayerWrapperDoesNotExist(
            "Player not found. Maybe it is a show and not a series?"
        )

    # Data are stored in a list, we extract it as bs4.Tag
    player_data: Tag = player_wrapper[0]
    data_entry: str = str(player_data.get("data-entry", ""))
    crologger.info("Getting data-entry values.")

    if not data_entry:
        err_msg = "Attribute data-entry not found."
        crologger.error(err_msg)
        raise DataEntryDoesNotExist(err_msg)

    json_data_entry = json.loads(data_entry)
    crologger.info("Parsing JSON.")

    uuid = json_data_entry.get("uuid", None)
    crologger.info("Getting UUID... %s", uuid)

    if not uuid:
        err_msg = "The program does not exist or other fatal error occured."
        crologger.error(err_msg)
        raise AudioUUIDDoesNotExist(err_msg)

    return uuid


def get_show_uuid(site_url: str, session: Session) -> str:
    """Return the show's UUID from the site URL."""

    crologger.info("Opening URL: %s", site_url)
    response = session.get(site_url, timeout=5)

    if response.status_code == 404:
        crologger.error("%s does not exist.", site_url)
        raise PageDoesNotExist("The page does not exist.")

    # Show UUID and other attributes are in the div with class b-detail
    soup = BeautifulSoup(response.text, "html.parser")
    div = soup.find("div", {"class": "b-detail"})

    crologger.info("Getting data-entry values.")

    data_entry: str = str(div.get("data-entry", ""))  # type: ignore

    if not data_entry:
        err_msg = "Attribute data-entry not found."
        crologger.error(err_msg)
        raise DataEntryDoesNotExist(err_msg)

    json_data_entry = json.loads(data_entry)
    crologger.info("Parsing JSON.")

    uuid = json_data_entry.get("show-uuid", None)
    crologger.info("Getting UUID... %s", uuid)

    if not uuid:
        err_msg = "Show does not exist or other fatal error occured."
        crologger.error(err_msg)
        raise ShowUUIDDoesNotExist(err_msg)

    return uuid


def get_attributes(uuid: str, session: Session) -> dict:
    """Returns the attributes of the episode with the given UUID."""
    response = session.get(API_SERVER + "/episodes/" + uuid, timeout=5)

    if not response.json():
        err_msg = "API server sent an empty answer. Attributes are not known."
        crologger.error(err_msg)
        raise AttributeError(err_msg)
    if not response.json().get("data"):
        err_msg = "Data is empty, cannot find 'attributes' (uuid: %s)" % uuid
        crologger.error(err_msg)

        return {}

    return response.json().get("data").get("attributes")


def get_audio_link_of_preferred_format(attrs: dict) -> str | None:
    """Searches for an audio link of preferred audio format. If not found, returns None."""
    audio_links = attrs.get("audioLinks")

    if not audio_links:
        return None

    audio_variants = [link.get("variant") for link in audio_links]
    audio_formats = {link.get("variant"): link.get("url") for link in audio_links}

    if PREFERRED_AUDIO_FORMAT.value not in audio_variants:
        audio_format = get_preferred_audio_format(audio_variants)
    else:
        audio_format = PREFERRED_AUDIO_FORMAT

    return audio_formats.get(audio_format.value) if audio_format else None


def get_js_value_from_url(site_url: str, jsvar: str, session: Session) -> str | None:
    """Returns the value of a JavaScript variable from the given URL."""
    response = session.get(site_url)
    err_msg = f"Variable {jsvar} was not found."

    if response.status_code == 200:
        soup = BeautifulSoup(response.text, "html.parser")
        script_tags = soup.find_all("script")

        for script in script_tags:
            if jsvar in script.text:
                pattern = re.compile(rf'"{jsvar}":"([^"]+)"')
                match = pattern.search(script.text)

                if match:
                    return match.group(1)

        crologger.error(err_msg)
        return None

    crologger.error("Server error.")
    return None


def is_series(url: str, session: Session) -> bool:
    """Returns True, if the page contains a series."""
    return get_js_value_from_url(url, "siteEntityBundle", session) == "serial"


def get_series_id(site_url: str, session: Session) -> str | None:
    """Returns series ID, based on its URL"""

    return get_js_value_from_url(site_url, "contentId", session)


def is_show(url: str, session: Session) -> bool:
    """Returns True,  if the page contains a show."""
    return get_js_value_from_url(url, "siteEntityBundle", session) == "show"


# "siteEntityBundle":"serialPart"



import datetime
import os
import re
import itertools

from pathlib import Path
from typing import TYPE_CHECKING, Optional
from zoneinfo import ZoneInfo

from crodl.settings import AudioFormat, DOWNLOAD_PATH
from crodl.tools.logger import crologger

if TYPE_CHECKING:
    from crodl.program.audiowork import AudioWork


def get_m4a_url(audio_link: str) -> str:
    """Returns the URL of a m4a dir with chunks (m4s or acc)

    Args:
        audio_link (str) -- URL of the audiowork (represented by 'manifest.mpd' or 'playlist.m3u8')

    Returns:
        Optional (str) -- URL of the m4a dir or None if neither pattern is found.

    """
    if "manifest.mpd" in audio_link:
        return audio_link.replace("/manifest.mpd", "")
    if "playlist.m3u8" in audio_link:
        return audio_link.replace("/playlist.m3u8", "")

    raise ValueError("Audio link is not valid.")


def partial_sums(nlist: list[int]) -> list[int]:
    """Returns a partial sums list of input list elements.

    Example:
        numbers = [6, 5, 6, 5, 6, 3],
        result = [6, 11, 17, 22, 28, 31]
    """
    return list(itertools.accumulate(nlist))


def get_preferred_audio_format(audio_variants: list[str]) -> AudioFormat | None:
    """
    Get the preferred audio format from a list of audio variants.

    Args:
        audio_variants (list[str]): A list of audio variants to choose from.

    Returns:
        AudioFormat: The preferred audio format, or None if none of the variants
        match the available formats.
    """
    for audio_format in AudioFormat:
        if audio_format.value in audio_variants:
            return audio_format

    return None


def process_audiowork_title(title: str, prefix: str | None = None) -> str:
    """Process the audiowork title to get a valid filename.
    Getting rid of a colon and a slash, replacing them with a dash."""
    if ":" in title:
        title = title.replace(":", " -")

    if "/" in title:
        title = title.replace("/", "-")

    if prefix:
        title = f"{prefix} - {title}"
    return title[:250]


def audio_segment_sort(filename) -> int | float:
    # Extract the numeric part from the filename
    # Making sure the format / suffix is not used
    filename = filename.split(".")[0]
    match = re.search(r"\d+", filename)
    if match:
        return int(match.group())
    return float("inf")  # Return infinity for non-numeric filenames


def simplify_audio_name(manifest_id: str, audio_name: str) -> str:
    """Renames complicated names by CRo.

    Example:
        segment_ctaudio_ridp0aa0br193031_cs80640000_mpd.m4s
        -> 80640000.m4s

        where the new name is basically a time stamp of DASH stream segment.

        "segment_ctaudio_rid" is a prefix,
        "p0aa0br193031" is a manifest_id

    """

    prefix = "segment_ctaudio_rid"

    # Removing characters from the left
    if "cinit" not in audio_name:
        audio_name = audio_name.replace(prefix + manifest_id + "_cs", "")
    else:
        audio_name = audio_name.replace(prefix + manifest_id + "_", "")

    # Removing characters from the right
    audio_name = audio_name.replace("_mpd", "").strip()

    return audio_name


def create_dir_if_does_not_exist(path: Path) -> None:
    if not path:
        raise ValueError("Path cannot be empty.")
    if not Path(path).exists():
        crologger.info("Creating a dir %s", path)
        os.makedirs(path)


def day_month_year(json_time: str) -> str:
    date_obj = datetime.datetime.strptime(json_time, "%Y-%m-%dT%H:%M:%S%z")
    return date_obj.strftime("%d-%m-%Y")


def title_with_part(title: str, part: int | str | None = None) -> str:
    if not title:
        raise ValueError("Title cannot be empty.")

    if part:
        if isinstance(part, int):
            part = str(part)
        if part.isnumeric():
            return f"{part}-{title}"
        raise ValueError("Part must be numeric.")

    return title


class HMS:
    def __init__(self, secs: int):
        self.secs = secs

    def __str__(self) -> str:
        hrs, mins, secs = self.hms()
        return f"{hrs:02d}:{mins:02d}:{secs:02d}"

    def hms(self) -> tuple[int, int, int]:
        """Transform seconds into hours, minutes and seconds."""
        hrs = self.secs // 3600
        mins = (self.secs % 3600) // 60
        secs = self.secs % 60
        return hrs, mins, secs


def file_size(size_in_bytes: int):
    if not isinstance(size_in_bytes, int):
        return size_in_bytes
    if size_in_bytes < 1024:
        return f"{size_in_bytes} B"
    if size_in_bytes < 1024**2:
        return f"{size_in_bytes / 1024:.2f} KB"

    return f"{size_in_bytes / 1024**2:.2f} MB"


def parse_date_from_json(json_time: str) -> tuple[str, str] | None:
    try:
        dt = datetime.datetime.fromisoformat(json_time)
    except ValueError:
        print("Cannot process the date.")
        dt = None
    return (dt.strftime("%d.%m.%Y"), dt.strftime("%H:%M")) if dt else None


def create_a_file_if_does_not_exist(path: Path, msg: Optional[str] = "") -> None:
    """Creates an empty file if it does not exist"""
    if not os.path.exists(path):
        crologger.info("Creatiung %s", path)

        try:
            with open(path, "w", encoding="utf-8") as f:
                if msg:
                    f.write(msg)
        except IOError as err:
            crologger.error(err)
            raise err


def not_available_yet(audiowork: "AudioWork") -> str:
    msg = "Datum a čas dostupnosti díla nebyl nalezen."
    err_msg = "Cannot find the release date and time. Data missing."

    if not audiowork.since:
        crologger.error(err_msg)
        return msg

    parsed_since = parse_date_from_json(audiowork.since)

    if not parsed_since:
        crologger.error(err_msg)
        return msg

    since = datetime.datetime.fromisoformat(audiowork.since)
    now = datetime.datetime.now(ZoneInfo("Europe/Prague"))

    if now < since:
        msg = f"Epizoda bude uvedena {parsed_since[0]} v {parsed_since[1]}."
        # create_a_file_if_not_exists(audiowork.audiowork_root + "/.series")
    else:
        crologger.info(
            f"Aired: {parsed_since[0]} at {parsed_since[1]}. Copryright license might have expired."
        )
        msg = f"Epizoda byla uvedena {parsed_since[0]} v {parsed_since[1]}. Možná vypršela práva."

    return msg


def unfinished_series() -> list[str]:
    series = []
    for root, _, files in os.walk(DOWNLOAD_PATH):
        if ".series" in files:
            series.append(root)
    return series


def shorten_title(title: str, length_limit: int) -> str:
    """Shortens the title if it is too long for console."""
    if len(title) > length_limit:
        return title[: length_limit - 3] + "..."
    return title


def get_audioformat_enum_by_value(val: str) -> AudioFormat | None:
    """Searches the enum AudioFormat by its value.

    Args:
        val (str): AudioFormat value (e.g. 'dash')

    Returns:
        AudioFormat: The AudioFormat enum (e.g. 'AudioFormat.DASH')
    """
    for key in AudioFormat:
        if key.value == val:
            return key
    return None


def remove_html_tags(text: str | None) -> str | None:
    """Removes HTML tags from text."""
    html_pattern = re.compile("<.*?>")
    if text:
        clean_text = re.sub(html_pattern, "", text)
        clean_text = clean_text.replace("&nbsp;", " ")
        return clean_text
    return None



import subprocess
from pathlib import Path
from dataclasses import dataclass, field
import shutil

from yaspin import yaspin

from crodl.settings import DOWNLOAD_PATH, SEGMENTS_SUBDIR, SUPPORTED_AUDIO_FORMATS
from crodl.tools.logger import crologger
from crodl.streams.utils import create_dir_if_does_not_exist, process_audiowork_title


@dataclass
class AudioParts:
    url: str
    audio_title: str
    audiowork_dir: Path | None = field(default=None)
    segments_path: Path | None = field(default=None)
    segments: bool = True

    def __post_init__(self):
        """After class init, this method will create audiowork_dir if it doesn't exist."""
        crologger.info(self.audio_title)

        if not self.audiowork_dir:
            self.audiowork_dir = DOWNLOAD_PATH / process_audiowork_title(
                self.audio_title
            )
            crologger.info("Creating an attribute %s", self.audiowork_dir)

        if not isinstance(self.audiowork_dir, Path):
            self.audiowork_dir = Path(self.audiowork_dir)

        if self.segments:
            if not self.segments_path:
                self.segments_path = Path(self.audiowork_dir) / SEGMENTS_SUBDIR

            create_dir_if_does_not_exist(self.segments_path)

    def _merge_chunks(self, audio_format: str) -> None:
        """
        Merges chunks of audio files into a final audiowork using ffmpeg.
        """
        if audio_format not in SUPPORTED_AUDIO_FORMATS:
            raise ValueError(f"Format '{audio_format}' is not supported!")

        subdirectory = f"{self.segments_path}"
        crologger.info("Merging files...")
        command = [
            "ffmpeg",
            "-i",
            "concatf:list.txt",
            "-c",
            "copy",
            f"../{process_audiowork_title(self.audio_title)}.{audio_format}",
            "-loglevel",
            "quiet",
        ]

        with yaspin(text=f"Ukládám {self.audio_title}", color="yellow") as spinner:
            subprocess.run(command, cwd=subdirectory, check=True)
            spinner.ok("✔️")

    def _purge_chunks_dir(self) -> None:  # pragma: no cover
        """Deletes tmp directory with audio chunks."""
        if not self.segments_path:
            raise ValueError("segments_path was not set")

        if Path.exists(self.segments_path):
            crologger.info("Deleting .chunks...")
            shutil.rmtree(self.segments_path)



import os
import unittest
import tempfile
from pathlib import Path

from crodl.settings import AudioFormat
from crodl.streams.utils import (
    day_month_year,
    get_m4a_url,
    audio_segment_sort,
    get_preferred_audio_format,
    partial_sums,
    process_audiowork_title,
    simplify_audio_name,
    create_dir_if_does_not_exist,
    title_with_part,
)


class TestGetM4aUrl(unittest.TestCase):
    def test_manifest_mpd(self):
        audio_link = "https://example.com/manifest.mpd"
        expected_output = "https://example.com"
        self.assertEqual(get_m4a_url(audio_link), expected_output)

    def test_playlist_m3u8(self):
        audio_link = "https://example.com/playlist.m3u8"
        expected_output = "https://example.com"
        self.assertEqual(get_m4a_url(audio_link), expected_output)

    def test_neither_pattern(self):
        audio_link = "https://example.com/some_other_url"
        with self.assertRaises(ValueError):
            get_m4a_url(audio_link)

    def test_empty_audio_link(self):
        audio_link = ""
        with self.assertRaises(ValueError):
            get_m4a_url(audio_link)


class TestPartialSums(unittest.TestCase):
    def test_empty_list(self):
        self.assertEqual(partial_sums([]), [])

    def test_single_element_list(self):
        self.assertEqual(partial_sums([5]), [5])

    def test_multiple_elements_list(self):
        self.assertEqual(partial_sums([6, 5, 6, 5, 6, 3]), [6, 11, 17, 22, 28, 31])


class TestGetPreferredAudioFormat(unittest.TestCase):
    def test_preferred_audio_format_first_variant(self):
        audio_variants = ["mp3", "hls", "dash"]
        result = get_preferred_audio_format(audio_variants)
        self.assertEqual(result, AudioFormat.MP3)

    def test_preferred_audio_format_last_variant(self):
        audio_variants = ["wma", "ogg", "dash"]
        result = get_preferred_audio_format(audio_variants)
        self.assertEqual(result, AudioFormat.DASH)

    def test_no_matching_audio_format(self):
        audio_variants = ["wma", "ogg", "flac"]
        result = get_preferred_audio_format(audio_variants)
        self.assertIsNone(result)


class TestProcessAudioworkTitle(unittest.TestCase):
    def test_title_contains_colon(self):
        title = "Sample: Title"
        processed_title = process_audiowork_title(title)
        self.assertEqual(processed_title, "Sample - Title")

    def test_title_does_not_contain_colon(self):
        title = "Sample Title"
        processed_title = process_audiowork_title(title)
        self.assertEqual(processed_title, "Sample Title")

    def test_title_contains_colon_and_prefix(self):
        title = "Sample: Title"
        processed_title = process_audiowork_title(title, prefix="01")
        self.assertEqual(processed_title, "01 - Sample - Title")


class TestAudioSegmentSort(unittest.TestCase):
    def test_numeric_filename(self):
        result = audio_segment_sort("audio123.mp3")
        self.assertEqual(result, 123)

    def test_non_numeric_filename(self):
        result = audio_segment_sort("audio.mp3")
        self.assertEqual(result, float("inf"))


class TestSimplifyAudioName(unittest.TestCase):
    """
    segment_ctaudio_ridp0aa0br193031_cs80640000_mpd.m4s
    prefix: segment_ctaudio_rid
    manifest_id: p0aa0br193031
    segment_time: 80640000
    """

    def test_no_cinit_in_audio_name(self):
        manifest_id = "p0aa0br193031"
        audio_name = "segment_ctaudio_ridp0aa0br193031_cs80640000_mpd.m4s"
        expected_output = "80640000.m4s"
        self.assertEqual(simplify_audio_name(manifest_id, audio_name), expected_output)

    def test_cinit_in_audio_name(self):
        """segment_ctaudio_ridp0aa0br193031_cinit_mpd.m4s"""
        manifest_id = "p0aa0br193031"
        audio_name = "segment_ctaudio_ridp0aa0br193031_cinit_mpd.m4s"
        expected_output = "cinit.m4s"
        self.assertEqual(simplify_audio_name(manifest_id, audio_name), expected_output)


class TestCreateDirIfDoesNotExist(unittest.TestCase):
    def test_create_dir_when_does_not_exist(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            dir_path = Path(tmp_dir) / "non_existent_dir"
            create_dir_if_does_not_exist(dir_path)
            self.assertTrue(os.path.exists(dir_path))

    def test_no_error_when_dir_already_exists(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            dir_path = Path(tmp_dir) / "existing_dir"
            os.makedirs(dir_path)
            create_dir_if_does_not_exist(dir_path)
            self.assertTrue(os.path.exists(dir_path))

    def test_error_when_path_is_invalid(self):
        with self.assertRaises(OSError):
            create_dir_if_does_not_exist(Path("/invalid/path/with/invalid:characters"))


class TestDayMonthYear(unittest.TestCase):
    def test_valid_date_string(self):
        json_time = "2022-07-25T14:30:00+0200"
        expected_output = "25-07-2022"
        self.assertEqual(day_month_year(json_time), expected_output)

    def test_invalid_date_string_format(self):
        json_time = "2022-07-25 14:30:00+0200"  # missing 'T' separator
        with self.assertRaises(ValueError):
            day_month_year(json_time)


class TestTitleWithPart(unittest.TestCase):
    def test_part_as_integer(self):
        title = "Example Title"
        part = 1
        expected = "1-Example Title"
        self.assertEqual(title_with_part(title, part), expected)

    def test_part_as_string(self):
        title = "Example Title"
        part = "one"

        with self.assertRaises(ValueError):
            title_with_part(title, part)

    def test_no_part(self):
        title = "Example Title"
        expected = "Example Title"
        self.assertEqual(title_with_part(title), expected)

    def test_empty_title(self):
        title = ""
        part = 1
        with self.assertRaises(ValueError):
            title_with_part(title, part)

    def test_title_with_special_chars(self):
        title = "Example Title with @#$%"
        part = 1
        expected = "1-Example Title with @#$%"
        self.assertEqual(title_with_part(title, part), expected)


if __name__ == "__main__":
    unittest.main()



from .exc import (
    DownloadError as DownloadError,
    DataEntryDoesNotExist as DataEntryDoesNotExist,
    PlayerWrapperDoesNotExist as PlayerWrapperDoesNotExist,
    AudioUUIDDoesNotExist as AudioUUIDDoesNotExist,
    ShowUUIDDoesNotExist as ShowUUIDDoesNotExist,
    PageDoesNotExist as PageDoesNotExist,
    DataNotFound as DataNotFound,
)



import os
import unittest
from unittest.mock import patch, MagicMock
from io import BytesIO
import tempfile
from pathlib import Path

from crodl.streams.mp3 import MP3


class TestMP3Download(unittest.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.TemporaryDirectory()
        self.mp3 = MP3(
            url="http://example.com/audio.mp3",
            audiowork_dir=Path(self.temp_dir.name),
            audio_title="My Audio",
        )

    @patch("requests.get")
    def test_download_success(self, mock_get):
        mock_response = MagicMock()
        mock_response.headers = {"Content-Length": "1024"}
        mock_response.raw = BytesIO(b"fake audio data")
        mock_get.return_value.__enter__.return_value = mock_response

        self.mp3.download()

        # Check if the file was created in the temporary directory
        expected_file_path = os.path.join(self.temp_dir.name, "My Audio.mp3")  # type: ignore
        self.assertTrue(os.path.exists(expected_file_path))

    @patch("requests.get")
    def test_download_no_content_length(self, mock_get):
        with tempfile.TemporaryDirectory() as temp_dir:
            self.temp_dir = temp_dir
            mock_response = MagicMock()
            mock_response.headers = {}
            mock_get.return_value.__enter__.return_value = mock_response

            with self.assertRaises(ValueError):
                self.mp3.download()


if __name__ == "__main__":
    unittest.main()



class DownloadError(Exception):
    pass


class DataEntryDoesNotExist(Exception):
    pass


class PlayerWrapperDoesNotExist(Exception):
    pass


class AudioUUIDDoesNotExist(Exception):
    pass


class ShowUUIDDoesNotExist(Exception):
    pass


class PageDoesNotExist(Exception):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(message)


class DataNotFound(KeyError):
    pass



import os
import shutil
from pathlib import Path

import unittest
from unittest.mock import MagicMock, patch

import tempfile
import asyncio
import aiohttp

from crodl.streams.download import download_part, download_parts, DownloadError


class TestDownloadPart(unittest.IsolatedAsyncioTestCase):
    async def test_successful_download(self):
        url = "https://example.com/file.txt"

        with tempfile.TemporaryDirectory() as target_folder:
            if not isinstance(target_folder, Path):
                target_folder = Path(target_folder)

            async with aiohttp.ClientSession() as session:
                with patch("aiohttp.ClientSession.get") as mock_get:
                    mock_response = MagicMock()
                    mock_response.status = 200
                    mock_response.content.read = MagicMock(
                        return_value=asyncio.Future()
                    )
                    mock_response.content.read.return_value.set_result(b"file content")
                    mock_get.return_value.__aenter__.return_value = mock_response

                    await download_part(url, session, target_folder)

                    self.assertTrue(os.path.exists(target_folder / "file.txt"))

    async def test_failed_download(self):
        url = "https://example.com/file.txt"
        target_folder = Path("/tmp")
        session = aiohttp.ClientSession()

        with patch("aiohttp.ClientSession.get") as mock_get:
            mock_response = MagicMock(status=404)
            mock_get.return_value.__aenter__.return_value = mock_response

            with self.assertRaises(DownloadError):
                await download_part(url, session, target_folder)
            await session.close()

    async def test_invalid_url(self):
        url = "invalid_url"
        target_folder = Path("/tmp")
        session = aiohttp.ClientSession()

        try:
            with self.assertRaises(aiohttp.ClientError):
                await download_part(url, session, target_folder)
        finally:
            await session.close()

    async def test_non_existent_target_folder(self):
        url = "https://example.com/file.txt"
        # Create a temporary directory and then delete it to simulate a non-existent folder
        target_folder = Path(tempfile.mkdtemp())
        shutil.rmtree(target_folder)  # Remove the directory to simulate it not existing

        async with aiohttp.ClientSession() as session:
            with patch("aiohttp.ClientSession.get") as mock_get:
                mock_response = MagicMock()
                mock_response.status = 200
                mock_response.content.read = MagicMock(return_value=asyncio.Future())
                mock_response.content.read.return_value.set_result(b"file content")
                mock_get.return_value.__aenter__.return_value = mock_response

                # Attempt to download the file to a non-existent folder and catch the exception
                with self.assertRaises(FileNotFoundError):
                    await download_part(url, session, target_folder)


class TestDownloadParts(unittest.IsolatedAsyncioTestCase):
    async def test_download_multiple_parts(self):
        urls = [
            "https://example.com/file1.txt",
            "https://example.com/file2.txt",
            "https://example.com/file3.txt",
        ]

        with tempfile.TemporaryDirectory() as target_folder:
            if not isinstance(target_folder, Path):
                target_folder = Path(target_folder)

            print(f"Temporary directory created at: {target_folder}")

            async with aiohttp.ClientSession() as _:
                with patch("aiohttp.ClientSession.get") as mock_get:
                    # Create mock responses
                    mock_responses = []
                    for i in range(3):
                        mock_response = MagicMock()
                        mock_response.status = 200
                        mock_response.content.read = MagicMock(
                            return_value=asyncio.Future()
                        )
                        mock_response.content.read.return_value.set_result(
                            f"content of file {i + 1}".encode()
                        )

                        mock_response.__aenter__.return_value = mock_response
                        mock_response.__aexit__.return_value = None
                        mock_responses.append(mock_response)

                    mock_get.side_effect = mock_responses
                    await download_parts(urls, target_folder)

                    for i, url in enumerate(urls):
                        file_name = target_folder / os.path.basename(url)
                        self.assertTrue(
                            os.path.exists(file_name), f"{file_name} does not exist"
                        )


if __name__ == "__main__":
    unittest.main()



import sys
from urllib.parse import urlparse

import asyncclick as click
from rich import print

from crodl.program.audiowork import AudioWork
from crodl.program.series import Series
from crodl.program.show import Show
from crodl.tools.scrap import cro_session, is_series, is_show
from crodl.tools.logger import crologger
from crodl.settings import SUPPORTED_DOMAINS, AudioFormat

from . import __version__


def is_domain_supported(url: str) -> bool:
    """Checks whether the website with 'hidden' audio lies in a supported domain."""
    domain = urlparse(url).netloc

    if not domain:
        err = "The URL is invalid!"
        crologger.error(err)
        print(err)
        raise ValueError(err)

    return domain in SUPPORTED_DOMAINS


def get_version(ctx, param, value) -> None:
    if not value or ctx.resilient_parsing:
        return
    click.echo(__version__)
    ctx.exit()


FORMAT_OPTIONS = {
    "mp3": AudioFormat.MP3,
    "hls": AudioFormat.HLS,
    "dash": AudioFormat.DASH,
}


@click.command()
@click.argument("recording_url")
@click.option(
    "--stream-format",
    "-sf",
    type=click.Choice(list(FORMAT_OPTIONS.keys())),
    default="mp3",
    help="Formát audio streamu. (mp3, hls, dash)",
)
@click.option(
    "--version", is_flag=True, callback=get_version, expose_value=False, is_eager=True
)
async def main(recording_url: str, stream_format: str) -> None:
    if not is_domain_supported(recording_url):
        raise NotImplementedError(
            f"Unsuported domain: {urlparse(recording_url).netloc}"
        )

    show, series = False, False

    if is_show(recording_url, cro_session):
        show = True
        audiowork = Show(url=recording_url)
        print(f"[bold yellow]{audiowork.title}[/bold yellow]")
        print(f"Celkový počet dílů: {audiowork.episodes.count}")

        if audiowork.already_exists():
            print("[bold yellow]Pořad byl již celý stažen.[/bold yellow]")
            sys.exit(0)

    elif is_series(recording_url, cro_session):
        series = True
        audiowork = Series(url=recording_url)
        print(f"[bold yellow]{audiowork.title}[/bold yellow]")
        print(f"[blue]{audiowork.description}[/blue]\n")
        print(f"Stažené díly: {str(audiowork.downloaded_parts)} / {audiowork.parts}")

        if audiowork.already_exists():
            print("[bold yellow]Seriál byl již celý stažen.[/bold yellow]")
            sys.exit(0)
    else:
        audiowork = AudioWork(url=recording_url)
        audiowork.info()

    if show or series:
        ans = input("Pokračovat ve stahování? [a/n]  ")
        if ans in ("a", "A", "y", "Y"):
            await audiowork.download(audio_format=FORMAT_OPTIONS[stream_format])
        else:
            print("[bold magenta]OK, končím. :wave:[/bold magenta]")
            sys.exit(0)
    else:
        if audiowork.already_exists():
            print("[bold magenta]Soubor již existuje.[/bold magenta] :wave:")
            sys.exit(0)
        await audiowork.download(audio_format=FORMAT_OPTIONS[stream_format])



import unittest
from unittest import mock
from unittest.mock import patch

from crodl.settings import DOWNLOAD_PATH
from crodl.program.audiowork import AudioWork


class TestAudioWorkInit(unittest.TestCase):
    def test_both_url_and_uuid_provided(self):
        with self.assertRaises(ValueError):
            AudioWork(url="https://example.com", uuid="12345")

    def test_neither_url_nor_uuid_provided(self):
        with self.assertRaises(ValueError):
            AudioWork()

    @patch("crodl.program.audiowork.get_audio_uuid")
    @patch("crodl.program.audiowork.get_attributes")
    def test_only_url_provided(self, mock_get_attributes, mock_get_audio_uuid):
        mock_get_audio_uuid.return_value = "12345"
        mock_get_attributes.return_value = {
            "title": "Example Title",
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(url="https://example.com")
        self.assertEqual(audio_work.url, "https://example.com")
        self.assertEqual(audio_work.uuid, "12345")
        self.assertEqual(audio_work.title, "Example Title")

    @patch("crodl.program.audiowork.get_attributes")
    def test_only_uuid_provided(self, mock_get_attributes):
        mock_get_attributes.return_value = {
            "title": "Example Title",
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(uuid="12345")
        self.assertEqual(audio_work.uuid, "12345")
        self.assertEqual(audio_work.title, "Example Title")

    @patch("crodl.program.audiowork.get_audio_uuid")
    @patch("crodl.program.audiowork.get_attributes")
    def test_title_not_provided(self, mock_get_attributes, mock_get_audio_uuid):
        mock_get_audio_uuid.return_value = "12345"
        mock_get_attributes.return_value = {
            "title": "Test Title",
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(url="https://example.com")
        self.assertEqual(audio_work.title, "Test Title")
        self.assertEqual(audio_work.audiowork_dir, DOWNLOAD_PATH / "Test Title")

    @patch("crodl.program.audiowork.get_audio_uuid")
    @patch("crodl.program.audiowork.get_attributes")
    def test_title_provided(self, mock_get_attributes, mock_get_audio_uuid):
        mock_get_audio_uuid.return_value = "12345"
        mock_get_attributes.return_value = {
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(url="https://example.com", title="Test Title")
        self.assertEqual(audio_work.title, "Test Title")

    @patch("crodl.program.audiowork.get_audio_uuid")
    @patch("crodl.program.audiowork.get_attributes")
    def test_audiowork_dir_not_provided(self, mock_get_attributes, mock_get_audio_uuid):
        mock_get_audio_uuid.return_value = "12345"
        mock_get_attributes.return_value = {
            "title": "Test title",
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(url="https://example.com")
        self.assertEqual(audio_work.audiowork_dir, DOWNLOAD_PATH / "Test title")

    @patch("crodl.program.audiowork.get_audio_uuid")
    @patch("crodl.program.audiowork.get_attributes")
    def test_audiowork_dir_provided(self, mock_get_attributes, mock_get_audio_uuid):
        mock_get_audio_uuid.return_value = "12345"
        mock_get_attributes.return_value = {
            "title": "Test title",
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(
            url="https://example.com", audiowork_dir=DOWNLOAD_PATH / "TestTitle"
        )
        self.assertEqual(audio_work.audiowork_dir, DOWNLOAD_PATH / "TestTitle")

    @patch("crodl.program.audiowork.get_audio_uuid")
    @patch("crodl.program.audiowork.get_attributes")
    def test_series_and_show_not_provided(
        self, mock_get_attributes, mock_get_audio_uuid
    ):
        mock_get_audio_uuid.return_value = "12345"
        mock_get_attributes.return_value = {
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(url="https://example.com")
        self.assertFalse(audio_work.series)
        self.assertFalse(audio_work.show)
        mock_get_audio_uuid.assert_called_once_with(
            audio_work.url, mock.ANY
        )  # Use mock.ANY to match any argument
        mock_get_attributes.assert_called_once_with("12345", mock.ANY)


class TestAudioWorkLinks(unittest.TestCase):
    @patch("crodl.program.audiowork.get_attributes")
    def test_audio_links_present(self, mock_get_attributes):
        mock_get_attributes.return_value = {
            "audioLinks": [{"link": "test_link"}],
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(uuid="12345")
        self.assertEqual(audio_work.audio_links, [{"link": "test_link"}])

    @patch("crodl.program.audiowork.get_attributes")
    def test_audio_links_not_present(self, mock_get_attributes):
        mock_get_attributes.return_value = {
            "audioLinks": [],
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(uuid="12345")

        self.assertIsNone(audio_work.audio_formats)


class TestAudioVariants(unittest.TestCase):
    @patch("crodl.program.audiowork.get_attributes")
    def test_audio_links_is_none(self, mock_get_attributes):
        mock_get_attributes.return_value = {
            "audioLinks": None,
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(uuid="12345")
        self.assertIsNone(audio_work.audio_formats)

    @patch("crodl.program.audiowork.get_attributes")
    def test_audio_links_is_empty_list(self, mock_get_attributes):
        mock_get_attributes.return_value = {
            "audioLinks": [],
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(uuid="12345")
        self.assertIsNone(audio_work.audio_formats)

    @patch("crodl.program.audiowork.get_attributes")
    def test_audio_links_has_no_variant_key(self, mock_get_attributes):
        mock_get_attributes.return_value = {
            "audioLinks": [{"key": "value"}],
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(uuid="12345")
        self.assertIsNone(audio_work.audio_formats)

    @patch("crodl.program.audiowork.get_attributes")
    def test_audio_links_has_variant_key(self, mock_get_attributes):
        mock_get_attributes.return_value = {
            "audioLinks": [{"variant": "mp3"}, {"variant": "aac"}],
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(uuid="12345")
        self.assertEqual(audio_work.audio_formats, ["mp3", "aac"])

    @patch("crodl.program.audiowork.get_attributes")
    def test_audio_links_is_not_a_list(self, mock_get_attributes):
        mock_get_attributes.return_value = {
            "audioLinks": "not a list",
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(uuid="12345")
        self.assertIsNone(audio_work.audio_formats)


class TestAudioWorkFormats(unittest.TestCase):
    @patch("crodl.program.audiowork.get_attributes")
    def test_audio_formats_with_audio_links(self, mock_get_attributes):
        expected_audio_links = [
            {"variant": "aac", "url": "https://example.com/aac.mp4"},
            {"variant": "m4a", "url": "https://example.com/m4a.mp4"},
        ]
        mock_get_attributes.return_value = {
            "audioLinks": expected_audio_links,
            "since": "2024-08-14T18:05:00+02:00",
        }
        audio_work = AudioWork(uuid="12345")
        expected_result = {
            "aac": "https://example.com/aac.mp4",
            "m4a": "https://example.com/m4a.mp4",
        }
        self.assertEqual(audio_work.audio_formats_urls, expected_result)

    @patch("crodl.program.audiowork.get_attributes")
    def test_audio_links_without_audio_links(self, mock_get_attributes):
        mock_get_attributes.return_value = {
            "audioLinks": [],
            "since": "2024-08-14T18:05:00+02:00",
        }

        audio_work = AudioWork(uuid="12345")
        self.assertIsNone(audio_work.audio_formats)


if __name__ == "__main__":
    unittest.main()



"""Import all program modules."""

from crodl.program.audiowork import AudioWork
from crodl.program.series import Series
from crodl.program.show import Show

__all__ = ["AudioWork", "Series", "Show"]



import tempfile

import shutil
from pathlib import Path

import unittest
from unittest.mock import patch, PropertyMock, Mock

from bs4 import BeautifulSoup
from crodl.streams.dash import (
    DASH,
    get_m4s_segment_url,
    audio_segment_times,
    segments_info,
    segments_urls,
)


class TestGetM4sSegmentUrl(unittest.TestCase):
    def test_init_segment_url(self):
        audio_link = "https://example.com/audio.mp4/manifest.mpd"
        repr_id = "p0aa0br193031"
        segment_time = None
        init = True
        expected_url = "https://example.com/audio.mp4/segment_ctaudio_ridp0aa0br193031_cinit_mpd.m4s"
        self.assertEqual(
            get_m4s_segment_url(audio_link, repr_id, segment_time, init), expected_url
        )

    def test_non_init_segment_url_integer(self):
        audio_link = "https://example.com/audio.mp4/manifest.mpd"
        repr_id = "p0aa0br193031"
        segment_time = 10
        init = False
        expected_url = "https://example.com/audio.mp4/segment_ctaudio_ridp0aa0br193031_cs10_mpd.m4s"
        self.assertEqual(
            get_m4s_segment_url(audio_link, repr_id, segment_time, init), expected_url
        )

    def test_non_init_segment_url_string(self):
        audio_link = "https://example.com/audio.mp4/manifest.mpd"
        repr_id = "p0aa0br193031"
        segment_time = "20"
        init = False
        expected_url = "https://example.com/audio.mp4/segment_ctaudio_ridp0aa0br193031_cs20_mpd.m4s"
        self.assertEqual(
            get_m4s_segment_url(audio_link, repr_id, segment_time, init), expected_url
        )


class TestAudioSegmentTimes(unittest.TestCase):
    def test_no_s_tags(self):
        mpd_content = BeautifulSoup("<MPD></MPD>", "xml")
        segment_times = audio_segment_times(mpd_content)
        self.assertEqual(segment_times, [])  # Expect an empty list

    def test_multiple_s_tags(self):
        mpd_content = BeautifulSoup(
            '<MPD><S d="3" r="1"/><S d="5" r="2"/></MPD>', "xml"
        )
        segment_times = audio_segment_times(mpd_content)
        self.assertEqual(
            segment_times,
            [3, 3, 5, 5, 5],
        )  # Expect [3, 3, 5, 5, 5]

    def test_missing_r_attribute(self):
        mpd_content = BeautifulSoup('<MPD><S d="2"/><S d="4" r="3"/></MPD>', "xml")
        segment_times = audio_segment_times(mpd_content)
        self.assertEqual(segment_times, [2, 4, 4, 4, 4])  # Expect [2, 4, 4, 4, 4]


@patch("crodl.streams.dash.audio_segment_times")
class TestSegmentsInfo(unittest.TestCase):
    def test_representation_found(self, mock_audio_segment_times):
        manifest_content = '<MPD><Representation id="123"></Representation></MPD>'
        mock_audio_segment_times.return_value = [10, 20, 30]
        expected_result = ([10, 30, 60], "123")
        self.assertEqual(segments_info(manifest_content), expected_result)

    def test_representation_not_found(self, mock_audio_segment_times):
        manifest_content = "<MPD></MPD>"
        mock_audio_segment_times.return_value = []
        with self.assertRaises(KeyError):
            segments_info(manifest_content)


class TestSegmentsUrls(unittest.TestCase):
    @patch("crodl.streams.dash.get_m4s_segment_url")  # Correct the patch target
    @patch.object(DASH, "content", new_callable=PropertyMock)
    def test_segments_urls(self, mock_content, mock_get_m4s_segment_url):
        manifest_content = """
        <MPD xmlns="urn:mpeg:dash:schema:mpd:2011" type="static" mediaPresentationDuration="PT10M">
            <Period>
                <AdaptationSet>
                    <SegmentTemplate timescale="1000" initialization="init.m4s">
                        <SegmentTimeline>
                            <S d="480240"/>
                            <S d="960528"/>
                            <S d="1920000"/>
                        </SegmentTimeline>
                    </SegmentTemplate>
                    <Representation id="09876abc" codecs="mp4a.40.2" audioSamplingRate="48000" bandwidth="128005">
                    </Representation>
                </AdaptationSet>
            </Period>
        </MPD>
        """
        mock_content.return_value = manifest_content

        def mock_get_url(audio_link, repr_id, segment_time=None, init=False):
            assert audio_link
            if init:
                return (
                    f"https://example.com/stream/xyz.m4a/audio_{repr_id}_cinit_mpd.m4s"
                )
            return f"https://example.com/stream/xyz.m4a/audio_{repr_id}_cs{segment_time}_mpd.m4s"

        mock_get_m4s_segment_url.side_effect = mock_get_url

        manifest_mock = Mock(spec=DASH)
        manifest_mock.url = "https://example.com/stream/xyz.m4a/manifest.mpd"
        manifest_mock.segments_path = "/some/path"

        manifest_mock.content = mock_content.return_value

        result = segments_urls(manifest_mock)

        expected_urls = [
            "https://example.com/stream/xyz.m4a/audio_09876abc_cinit_mpd.m4s",  # initial segment
            "https://example.com/stream/xyz.m4a/audio_09876abc_cs0_mpd.m4s",  # zeroth segment
            "https://example.com/stream/xyz.m4a/audio_09876abc_cs480240_mpd.m4s",  # first segment
            "https://example.com/stream/xyz.m4a/audio_09876abc_cs1440768_mpd.m4s",
        ]

        self.assertEqual(result, expected_urls)


class TestManifest(unittest.TestCase):
    @patch("crodl.streams.utils.get_m4a_url")
    @patch("crodl.tools.scrap.cro_session.get")
    def test_get_manifest(self, mock_get, mock_get_m4a_url):
        # Test case: fetching manifest and saving it locally
        mock_get_m4a_url.return_value = "https://example.com/test.mp4"
        mock_get.return_value.content = b'{"manifest": "test"}'

        manifest = DASH("https://example.com/test.mp4/manifest.mpd", "Some Title")
        manifest._get_manifest()  # pylint: disable=protected-access

        mock_get.assert_called_once_with(
            "https://example.com/test.mp4/manifest.mpd", timeout=5
        )
        with open(f"{manifest.segments_path}/manifest.mpd", "rb") as f:
            content = f.read()
        self.assertEqual(content, b'{"manifest": "test"}')


class TestManifestContent(unittest.TestCase):
    def test_content_exists(self):
        # Create a mock manifest file
        segments_path = Path(tempfile.mkdtemp())
        manifest_file = segments_path / "manifest.mpd"

        with open(manifest_file, "w", encoding="utf-8") as f:
            f.write("Mock manifest content")

        # Initialize Manifest with a mock URL
        url = "http://example.com/mp4/manifest.mpd"
        manifest = DASH(url, "Some Title")
        manifest.segments_path = segments_path

        # Test that the content property returns the manifest content
        self.assertEqual(manifest.content, "Mock manifest content")
        shutil.rmtree(segments_path)


@patch("crodl.streams.dash.segments_info")
class TestManifestIdProperty(unittest.TestCase):
    def setUp(self):
        self.manifest = DASH("http://example.com/mp4/manifest.mpd", "Some Title")

    def test_id_property_returns_correct_value(self, mock_segments_info):
        mock_segments_info.return_value = (None, "12345")
        self.assertEqual(self.manifest.id, "12345")

    def test_id_property_raises_exception_on_invalid_value(self, mock_segments_info):
        mock_segments_info.return_value = (None, None)
        with self.assertRaises(ValueError):
            self.manifest.id


class TestSegmentUrls(unittest.TestCase):
    def setUp(self):
        self.manifest = DASH("http://example.com/mp4/manifest.mpd", "Some Title")

    def test_segment_urls_returns_list_of_strings(self):
        with patch("crodl.streams.dash.segments_urls") as mock_segments_urls:
            mock_segments_urls.return_value = ["url1", "url2"]
            segment_urls = self.manifest.segment_urls
            self.assertIsInstance(segment_urls, list)
            self.assertEqual(len(segment_urls), 2)
            self.assertEqual(segment_urls[0], "url1")
            self.assertEqual(segment_urls[1], "url2")

    def test_segment_urls_calls_segments_urls(self):
        with patch("crodl.streams.dash.segments_urls") as mock_segments_urls:
            self.manifest.segment_urls
            mock_segments_urls.assert_called_once_with(self.manifest)

    def test_segment_urls_returns_empty_list(self):
        with patch("crodl.streams.dash.segments_urls") as mock_segments_urls:
            mock_segments_urls.return_value = []
            segment_urls = self.manifest.segment_urls
            self.assertEqual(segment_urls, [])

    def test_segment_urls_raises_exception(self):
        with patch("crodl.streams.dash.segments_urls") as mock_segments_urls:
            mock_segments_urls.side_effect = Exception("Test exception")
            with self.assertRaises(Exception):
                self.manifest.segment_urls


@patch("os.listdir")
class TestManifestCreateListTxt(unittest.TestCase):
    def setUp(self):
        self.manifest = DASH("http://example.com/mp4/manifest.mpd", "Some Title")
        self.temp_dir = tempfile.TemporaryDirectory()
        self.manifest.segments_path = Path(self.temp_dir.name)

    def tearDown(self):
        self.temp_dir.cleanup()

    @patch.object(DASH, "id", new_callable=PropertyMock)
    def test_no_segment_files(self, mock_listdir, mock_id):
        mock_id.return_value = "abc123"
        mock_listdir.return_value = []
        with self.assertRaises(ValueError):
            self.manifest.create_list_txt()

    @patch("crodl.streams.utils.simplify_audio_name")
    @patch("crodl.streams.utils.audio_segment_sort")
    @patch.object(DASH, "id", new_callable=PropertyMock)
    def test_create_list_txt(
        self, mock_id, mock_audio_segment_sort, mock_simplify_audio_name, mock_listdir
    ):
        mock_id.return_value = "abc123"
        mock_listdir.return_value = [
            "cinit.m4s",
            "segment_1.m4s",
            "segment_2.m4s",
            "segment_3.m4s",
        ]
        mock_simplify_audio_name.side_effect = [
            "cinit.m4s",
            "segment_1.m4s",
            "segment_2.m4s",
            "segment_3.m4s",
        ]
        mock_audio_segment_sort.return_value = 1

        self.manifest.create_list_txt()

        with open(
            f"{self.manifest.segments_path}/list.txt", "r", encoding="utf-8"
        ) as f:
            contents = f.read()
            self.assertEqual(
                contents, "cinit.m4s\nsegment_1.m4s\nsegment_2.m4s\nsegment_3.m4s\n"
            )


if __name__ == "__main__":
    unittest.main()



import os
from dataclasses import dataclass
from pathlib import Path

from bs4 import BeautifulSoup
from yaspin import yaspin

from crodl.streams.audioparts import AudioParts
from crodl.streams.download import download_parts
from crodl.streams.utils import (
    audio_segment_sort,
    get_m4a_url,
    partial_sums,
    simplify_audio_name,
)


def get_m4s_segment_url(
    audio_link: str,
    repr_id: str,
    segment_time: str | int | None = None,
    init: bool = False,
) -> str:
    """
    Get the segment URL for a given audio link and representation ID.

    Args:
        audio_link (str): The URL of the audio.
        repr_id (str): The representation ID.
        segment_time (str | int | None, optional): The segment time. Defaults to None.
        init (bool, optional): Whether it is an initial segment. Defaults to False.

    Returns:
        str | None: The segment URL if it exists, else None.
    """
    m4a_url = get_m4a_url(audio_link)

    if init:
        return f"{m4a_url}/segment_ctaudio_rid{repr_id}_cinit_mpd.m4s"

    return f"{m4a_url}/segment_ctaudio_rid{repr_id}_cs{segment_time}_mpd.m4s"


def audio_segment_times(mpd_content: BeautifulSoup) -> list[int]:
    """
    Return a list of stream segment times extracted from the MPD content.

    Args:
        mpd_content (BeautifulSoup): The parsed MPD content.

    Returns:
        list[int]: A list of segment times.
    """

    segment_times = []
    for s_tag in mpd_content.find_all("S"):
        duration = int(s_tag["d"])  # type: ignore

        # Default to 0 if 'r' attribute is not present
        repeat = int(s_tag.get("r", 0))  # type: ignore

        # Repeat 'r' times
        segment_times.extend([duration] * (repeat + 1))  # type: ignore

    return segment_times


def segments_info(manifest_content: str) -> tuple[list[int], str]:
    """
    Retrieve information about segments from the provided MPD Manifest file.

    Args:
        manifest_content (str): MPD file content.

    Returns:
        list[str]: A list containing the partial sums of d_values and the representation ID.
    """

    soup = BeautifulSoup(manifest_content, "xml")

    representation = soup.find("Representation", id=True)

    if not representation:
        raise KeyError("Block 'representation' not found.")

    representation_id = representation.get("id")  # type: ignore
    d_values = audio_segment_times(soup)

    return partial_sums(d_values), str(representation_id)


def segments_urls(manifest: "DASH") -> list[str]:
    """
    Generate a list of segment URLs for the given MPD URL.

    Args:
        manifest_content (str): The content of the MPD file.

    Returns:
        list[str]: A list of segment URLs.

    Example:
        [
        'https://example.com/stream/xyz.m4a/audio_09876abc_cinit_mpd.m4s', # initial segment
        'https://example.com/stream/xyz.m4a/audio_09876abc_cs0_mpd.m4s', # zeroth segment
        'https://example.com/stream/xyz.m4a/audio_09876abc_cs480240_mpd.m4s', # first segment
        'https://example.com/stream/xyz.m4a/audio_09876abc_cs960528_mpd.m4s',
        'https://example.com/stream/xyz.m4a/audio_09876abc_cs1439760_mpd.m4s',
        'https://example.com/stream/xyz.m4a/audio_09876abc_cs1920000_mpd.m4s',
        ...
        ]
    """
    segment_times, id_ = segments_info(manifest.content)
    init_chunk = get_m4s_segment_url(manifest.url, id_, init=True)
    zeroth_chunk = get_m4s_segment_url(manifest.url, id_, 0)

    # The last segment is not included in the list, according to DASH specification.
    return (
        [init_chunk]
        + [zeroth_chunk]
        + [
            get_m4s_segment_url(manifest.url, id_, segment_time)
            for segment_time in segment_times
        ][:-1]
    )


@dataclass
class DASH(AudioParts):
    """Processes a DASH stream using its manifest file."""

    @property
    def manifest_path(self) -> Path:  # pragma: no cover
        """Path to the manifest file"""
        if not self.segments_path:
            raise ValueError("Segments Path is not set!")
        return self.segments_path / "manifest.mpd"

    def _get_manifest(self) -> None:
        """Fetches the manifest and saves it to a file locally."""
        from crodl.tools.scrap import cro_session

        mp4_url = get_m4a_url(self.url)
        manifest_url = mp4_url + "/manifest.mpd"
        _manifest = cro_session.get(manifest_url, timeout=5)

        with open(self.manifest_path, "wb") as f:
            f.write(_manifest.content)

    @property
    def content(self) -> str:  # pragma: no cover
        """Manifest content"""
        if not os.path.isfile(self.manifest_path):
            self._get_manifest()

        with open(self.manifest_path, "r", encoding="utf-8") as f:
            return f.read()

    @property
    def id(self) -> str:
        """Audiowork ID in manifest.mpd"""
        repr_id = segments_info(self.content)[1]
        if repr_id and isinstance(repr_id, str):
            return repr_id

        raise ValueError("Representation ID not found.")

    @property
    def segment_urls(self) -> list[str]:
        """Return segment urls (urls of all audio segments)"""
        return segments_urls(self)

    def create_list_txt(self) -> None:
        """Creates a sorted list of all audio segment names (processed) and saves it to list.txt"""
        segment_names = list(os.listdir(self.segments_path))

        if not segment_names or all(
            not segment.endswith(".m4s") for segment in segment_names
        ):
            raise ValueError("Cannot create an audio file. Segments not found.")

        simplified_names = [
            simplify_audio_name(self.id, name)
            for name in segment_names
            if "init" not in name
        ]

        sorted_names = sorted(simplified_names, key=audio_segment_sort)
        sorted_names.insert(0, "cinit.m4s")

        # Write the segment names to list.txt
        if not self.segments_path:
            raise ValueError("Segments Path is not set!")
        if not isinstance(self.segments_path, Path):
            self.segments_path = Path(self.segments_path)

        list_path = self.segments_path / "list.txt"
        with list_path.open("w", encoding="utf-8") as segment:
            for name in sorted_names:
                segment.write(f"{name}\n")

    def rename_segments(self) -> None:  # pragma: no cover
        """Simplifies / renames audio segments."""
        for segment in os.listdir(self.segments_path):
            if segment.startswith("segment_"):
                new_name = simplify_audio_name(self.id, segment)

                if not self.segments_path:
                    raise ValueError("Segments Path is not set!")

                os.rename(
                    self.segments_path / segment,
                    self.segments_path / new_name,
                )

    async def download(self):  # pragma: no cover
        """
        Asynchronously downloads chunk files using the segment URLs
        and saves them to the specified segments path.

        After the download is completed, merging is run and, finally,
        the tmp dir with chunks is deleted.
        """
        if not self.segments_path:
            raise ValueError("Segments Path is not set!")

        with yaspin(text=f"Stahuji {self.audio_title}", color="yellow") as _:
            await download_parts(self.segment_urls, self.segments_path)

        with yaspin(text=f"Zpracovávám {self.audio_title}", color="yellow") as _:
            self.create_list_txt()
            self.rename_segments()

        self._merge_chunks("m4a")
        self._purge_chunks_dir()



from dataclasses import dataclass, field
import os

from typing import Optional
from pathlib import Path

from crodl.program.content import Content
from crodl.settings import (
    API_SERVER,
    DOWNLOAD_PATH,
    AudioFormat,
    PREFERRED_AUDIO_FORMAT,
    AUDIO_FORMATS,
)
from crodl.program.audiowork import AudioWork
from crodl.tools.logger import crologger
from crodl.streams.utils import create_dir_if_does_not_exist, title_with_part
from crodl.tools.scrap import (
    cro_session,
    get_audio_link_of_preferred_format,
    get_show_uuid,
)


@dataclass
class Attributes:
    title: str
    active: bool
    aired: bool
    description: str
    short_description: str


@dataclass
class Data:
    type: str
    id: str
    attributes: Attributes


# TODO: Episode class
# TODO: Support more attrs.


@dataclass
class Episodes:
    show_id: str
    count: int = field(init=False)
    data: list[dict] = field(init=False)

    def __post_init__(self):
        episodes_url = f"{API_SERVER}/shows/{self.show_id}/episodes"

        response = cro_session.get(episodes_url)
        response.raise_for_status()
        json_data = response.json()

        self.data = json_data["data"]
        self.count = json_data["meta"]["count"]

    @property
    def info(self) -> list[str]:
        info = []
        for _data in self.data:
            attrs = _data["attributes"]
            info.append(
                {
                    "uuid": _data["id"],
                    "title": attrs["title"],
                    "url": get_audio_link_of_preferred_format(attrs),
                    "since": attrs["since"],
                    "part": attrs["part"],
                }
            )

        return info

    # @property
    # def df(self) -> pds.DataFrame | str:
    #     dframe = pds.DataFrame(self.info)
    #     dframe = dframe.sort_values(
    #         ['since'],
    #         ascending=[
    #             False,
    #         ],
    #     )
    #     dframe = dframe.drop(['uuid', 'url'], axis=1).reset_index(drop=True)
    #     dframe.rename(columns={'title': 'Titul', 'since': 'Vysíláno', 'part': 'Díl'}, inplace=True)

    #     # Nahrazení NaN hodnot nulou
    #     dframe['Díl'] = dframe['Díl'].fillna(0)

    #     # Převod sloupce na integer
    #     dframe['Díl'] = dframe['Díl'].astype(int)
    #     # dframe = dframe[["Díl", "Titul", "Vysíláno"]]

    #     dframe = dframe.to_string(index=False)
    #     return dframe


@dataclass
class Show(Content):
    """
    Class for processing Shows by ČRo.
    Example: https://www.mujrozhlas.cz/dan-barta-nevinnosti-sveta

    In HTML, we look for show uuid -- found: 1a9044a7-18a2-32fe-870a-32ec9bf33c74
    Then we use an API call to get info on this show:
        https://api.mujrozhlas.cz/shows/1a9044a7-18a2-32fe-870a-32ec9bf33c74
    Corresponding show episodes are then taken from this API call:
        https://api.mujrozhlas.cz/shows/1a9044a7-18a2-32fe-870a-32ec9bf33c74/episodes
    """

    api_url: str = field(init=False)
    data: Data = field(init=False)
    episodes: Episodes = field(init=False)
    download_dir: Path = field(init=False)

    def __post_init__(self):
        show_id = get_show_uuid(self.url, cro_session)
        show_api_url = f"{API_SERVER}/shows/{show_id}"

        # Fetch the JSON data from Show URL
        response = cro_session.get(show_api_url)
        response.raise_for_status()  # Raise an exception for HTTP errors
        json_data = response.json()

        attributes = Attributes(
            title=json_data["data"]["attributes"]["title"],
            active=json_data["data"]["attributes"]["active"],
            aired=json_data["data"]["attributes"]["aired"],
            description=json_data["data"]["attributes"]["description"],
            short_description=json_data["data"]["attributes"]["shortDescription"],
        )

        data = Data(
            type=json_data["data"]["type"],
            id=json_data["data"]["id"],
            attributes=attributes,
        )

        self.title = json_data["data"]["attributes"]["title"]
        self.api_url = show_api_url
        self.data = data
        self.episodes = Episodes(show_id)
        self.download_dir = DOWNLOAD_PATH / self.title

    @property
    def downloaded_parts(self) -> int:
        crologger.info("Checking whether the show has been already downloaded...")
        if not os.path.isdir(self.download_dir):
            return False

        downloaded_parts = sum(
            1
            for file in os.listdir(self.download_dir)
            if file.lower().endswith(AUDIO_FORMATS)
        )

        crologger.info("Parts downloaded: %s", downloaded_parts)
        crologger.info("Total parts: %s", self.episodes.count)

        return downloaded_parts

    def already_exists(self) -> bool:
        return self.downloaded_parts == self.episodes.count

    async def download(
        self, audio_format: Optional[AudioFormat] = PREFERRED_AUDIO_FORMAT
    ) -> None:
        """Downloads all episodes of the series to their own subfolders."""
        create_dir_if_does_not_exist(self.download_dir)

        for episode in self.episodes.info:
            download_to = self.download_dir
            audio_work = AudioWork(
                uuid=episode.get("uuid"),  # type: ignore
                audiowork_dir=download_to,
                title=title_with_part(episode.get("title"), episode.get("part")),  # type: ignore
                since=episode.get("since"),  # type: ignore
                show=True,
            )

            await audio_work.download(audio_format)



import unittest
from crodl.main import is_domain_supported


class TestIsDomainSupported(unittest.TestCase):
    def test_supported_domain(self):
        url = "https://www.mujrozhlas.cz"
        self.assertTrue(is_domain_supported(url))

    def test_unsupported_domain(self):
        url = "https://vltava.rozhlas.cz"
        self.assertFalse(is_domain_supported(url))

    def test_empty_url(self):
        url = ""
        with self.assertRaises(ValueError):
            is_domain_supported(url)

    def test_invalid_url(self):
        url = " invalid url "
        with self.assertRaises(ValueError):
            is_domain_supported(url)


if __name__ == "__main__":
    unittest.main()



# tests/test_audioparts.py
import os
import unittest
import tempfile
from pathlib import Path
from unittest.mock import patch
from crodl.streams.audioparts import AudioParts
from crodl.streams.utils import create_dir_if_does_not_exist, process_audiowork_title


class TestAudioParts(unittest.TestCase):
    def setUp(self):
        self.title = "Some Title"
        self.url = "https://example.com"
        self.audio_parts = AudioParts(url=self.url, audio_title=self.title)

    def test_init(self):
        self.assertEqual(self.audio_parts.url, self.url)
        self.assertEqual(self.audio_parts.audio_title, self.title)

    def test_create_dir_if_does_not_exist_real_dir(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_dir = Path(tmp_dir)
            dir_path = tmp_dir / "test_dir"
            create_dir_if_does_not_exist(dir_path)
            self.assertTrue(os.path.exists(dir_path))


class TestMergeChunks(unittest.TestCase):
    def setUp(self):
        self.audio_parts = AudioParts("url", "audio_title")

    @patch("subprocess.run")
    def test_supported_audio_format_m4a(self, mock_subprocess_run):
        self.audio_parts._merge_chunks("m4a")  # pylint: disable=protected-access
        mock_subprocess_run.assert_called_once()

    @patch("subprocess.run")
    def test_supported_audio_format_aac(self, mock_subprocess_run):
        self.audio_parts._merge_chunks("aac")  # pylint: disable=protected-access
        mock_subprocess_run.assert_called_once()

    def test_unsupported_audio_format(self):
        with self.assertRaises(ValueError):
            self.audio_parts._merge_chunks("mp3")  # pylint: disable=protected-access

    @patch("subprocess.run")
    def test_subprocess_run_args(self, mock_subprocess_run):
        self.audio_parts._merge_chunks("m4a")  # pylint: disable=protected-access

        expected_command = [
            "ffmpeg",
            "-i",
            "concatf:list.txt",
            "-c",
            "copy",
            f"../{process_audiowork_title(self.audio_parts.audio_title)}.m4a",
            "-loglevel",
            "quiet",
        ]
        mock_subprocess_run.assert_called_once_with(
            expected_command, cwd=str(self.audio_parts.segments_path), check=True
        )


if __name__ == "__main__":
    unittest.main()



import sys
import os

from pathlib import Path
from dataclasses import dataclass, field

from typing import Optional

from crodl.program.content import Content
from crodl.settings import (
    DOWNLOAD_PATH,
    API_SERVER,
    PREFERRED_AUDIO_FORMAT,
    AudioFormat,
)
from crodl.tools.scrap import (
    cro_session,
    get_audio_link_of_preferred_format,
    get_series_id,
)
from crodl.tools.logger import crologger
from crodl.program.audiowork import AudioWork
from crodl.streams.utils import (
    create_a_file_if_does_not_exist,
    process_audiowork_title,
    create_dir_if_does_not_exist,
    remove_html_tags,
)


@dataclass
class Series(Content):
    id: str = field(init=False)
    download_dir: Optional[Path] = field(default=None)

    def __post_init__(self):
        """
        Method initializes various attributes of the object if they are not already set.
        """
        self.id = get_series_id(self.url, cro_session)  # type: ignore
        self.json = cro_session.get(f"{API_SERVER}/serials/{self.id}", timeout=5).json()

        if not self.json:
            crologger.error("Got an empty response. Series might not be available.")
            print("Seriál není dostupný. Zkuste akci opakovat později.")
            sys.exit(0)

        self._attrs = self.json["data"]["attributes"]

        self.title = self._attrs["title"]
        self.parts = int(self._attrs["totalParts"])

        crologger.info("Opening series %s", self.title)
        crologger.info("Series ID : %s", self.id)
        crologger.info("Episodes: %s", self.parts)

        if not self.is_playable:
            msg = f"Series {self.title} is not available."
            crologger.error(msg)
            print("Seriál není dostupný.")
            sys.exit(0)

        if not self.download_dir:
            self.download_dir = (
                DOWNLOAD_PATH / "Seriály" / process_audiowork_title(self.title)
            )

    def __str__(self) -> str:
        return f"<Series: {self.title}>"

    def __repr__(self) -> str:
        return f"<Series: {self.title}>"

    @property
    def description(self) -> str | None:
        """
        A property method that returns the description of the series.
        """

        if not self._attrs:
            return None

        if self._attrs.get("description"):
            # Remove HTML tags and return the description
            return remove_html_tags(self._attrs.get("description"))
        return None

    @property
    def is_playable(self) -> bool:
        if self._attrs:
            return self._attrs.get("playable") is True

        return False

    @property
    def _episodes_url(self) -> str:
        """
        A property method that returns the API URL for episodes based on the data provided.
        Return type: str
        """
        return (
            self.json.get("data")
            .get("relationships")  # type: ignore
            .get("episodes")
            .get("links")
            .get("related")
        )

    def _fetch_episodes(self) -> list[dict]:
        """
        A private method to fetch episodes using `cro_session` with a timeout of 5 seconds.

        Returns:
            A list of dictionaries representing the fetched episodes.
        """
        return cro_session.get(self._episodes_url, timeout=5).json().get("data")

    @property
    def episodes_data(self) -> list[dict]:
        """
        A property method that returns the episodes data as a list of dictionaries.
        """
        return self._fetch_episodes()

    @property
    def audio_formats(self) -> list[str | None]:
        episode = self.episodes_data[0]

        if not episode:
            return []

        audiolinks = episode["attributes"]["audioLinks"]

        formats = []
        for audiolink in audiolinks:
            formats.append(audiolink["variant"])
        return formats

    def list_all_series_episodes(self) -> list[str]:
        """
        Lists all series episodes and their details.

        Returns:
            A list of dictionaries containing episode information.
        """
        json_data = self.episodes_data

        all_parts = []

        for data in json_data:
            attrs = data.get("attributes")
            episode_num = attrs.get("part")  # type: ignore
            audio_link = get_audio_link_of_preferred_format(attrs)  # type: ignore

            all_parts.append(
                {
                    "uuid": data.get("id"),
                    "title": f"{episode_num}" + "-" + attrs.get("title"),  # type: ignore
                    "episode_num": episode_num,
                    "url": audio_link,
                    "since": attrs.get("since"),  # type: ignore
                }
            )

        return all_parts

    @property
    def downloaded_parts(self) -> int:
        crologger.info("Checking whether the series has been downloaded already...")
        if not self.download_dir:
            raise ValueError("Download dir is not set!")
        if not os.path.isdir(self.download_dir):
            return 0

        downloaded_parts = 0
        for part in range(1, self.parts + 1):
            if any(f.startswith(f"{part}-") for f in os.listdir(self.download_dir)):
                downloaded_parts += 1

        crologger.info("Parts downloaded: %s", downloaded_parts)
        crologger.info("Total parts: %s", self.parts)

        return downloaded_parts

    def already_exists(self) -> bool:
        return self.downloaded_parts == self.parts

    async def download(
        self, audio_format: Optional[AudioFormat] = PREFERRED_AUDIO_FORMAT
    ) -> None:
        """Downloads all series episodes."""
        if not self.download_dir:
            raise ValueError("Download dir is not set!")

        create_dir_if_does_not_exist(self.download_dir)
        create_a_file_if_does_not_exist(self.download_dir / ".series")

        for episode in self.list_all_series_episodes():
            download_to = self.download_dir
            audio_work = AudioWork(
                uuid=episode.get("uuid"),  # type: ignore
                audiowork_root=self.download_dir,
                audiowork_dir=download_to,
                title=episode.get("title"),  # type: ignore
                series=True,
                since=episode.get("since"),  # type: ignore
            )
            await audio_work.download(audio_format)



import time


class Timer:
    """
    Context Manager measuring time needed to run and finish a piece of code.

    Example:

        with Timer() as timer:
            my_code_goes_here

    Useful when finding a bottleneck in codes.
    """

    def __init__(self):
        self.start_time = 0

    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, *_):
        elapsed_time = time.time() - self.start_time
        print(f"Time elapsed: {elapsed_time} seconds")



from dataclasses import dataclass


import requests
from rich.progress import Progress

from crodl.streams.audioparts import AudioParts
from crodl.streams.utils import process_audiowork_title, shorten_title
from crodl.tools.logger import crologger


@dataclass
class MP3(AudioParts):
    """Process mp3 stream from CRo"""

    def download(self) -> None:
        """Download audiowork mp3 file"""
        file_url = self.url
        crologger.info("Downloading mp3: %s", file_url)

        with requests.get(file_url, timeout=4, stream=True) as resp:
            content_length = resp.headers.get("Content-Length")

            if not self.audiowork_dir:
                raise ValueError("self.audiowork_dir is not set.")

            if not content_length:
                raise ValueError("The file cannot be downloaded.")

            total_length = int(content_length)

            with Progress() as progress:
                task = progress.add_task(
                    shorten_title(self.audio_title, 20), total=total_length
                )

                audio_full_path = (
                    self.audiowork_dir
                    / f"{process_audiowork_title(self.audio_title)}.mp3"
                )
                with audio_full_path.open("wb") as output:
                    while True:
                        chunk = resp.raw.read(1024)
                        if not chunk:
                            break

                        output.write(chunk)
                        progress.update(task, advance=len(chunk))



import os
import asyncio
import aiohttp

from pathlib import Path

from crodl.tools.logger import crologger
from crodl.exceptions import DownloadError


async def download_part(url: str, session: aiohttp.ClientSession, target_folder: Path):
    """Download a single part (segment / chunk) of an audio stream."""
    file_name = target_folder / os.path.basename(url)

    async with session.get(url) as response:
        if response.status == 200:
            with open(file_name, "wb") as f:
                content = await response.content.read()
                f.write(content)

            crologger.info("Downloading %s... OK", url)
        else:
            crologger.error("Downloading %s... Error", url)
            raise DownloadError(
                f"Error whiel downloading {url}: HTTP {response.status}"
            )


async def download_parts(urls: list[str], target_folder: Path):
    """Download asynchronously audio parts (segments / chunks)."""
    os.makedirs(target_folder, exist_ok=True)  # Ensure the target folder exists

    async with aiohttp.ClientSession() as session:
        tasks = [download_part(url, session, target_folder) for url in urls]
        await asyncio.gather(*tasks)



import logging
import os

from crodl.settings import LOG_PATH


def create_log_dir():  # pragma: no cover
    if not os.path.exists(LOG_PATH):
        os.makedirs(LOG_PATH)


crologger = logging.getLogger(__name__)
crologger.setLevel(logging.DEBUG)

create_log_dir()

logfile = logging.FileHandler(os.path.join(LOG_PATH, "crodl.log"))
fileformat = logging.Formatter("%(asctime)s - %(levelname)s - %(module)s - %(message)s")
logfile.setFormatter(fileformat)
crologger.addHandler(logfile)



import os

from typing import Optional
from rich import print

from crodl.streams import DASH
from crodl.streams import HLS
from crodl.streams import MP3
from crodl.streams.utils import (
    HMS,
    file_size,
    get_preferred_audio_format,
    not_available_yet,
    process_audiowork_title,
    create_dir_if_does_not_exist,
    remove_html_tags,
)
from crodl.tools.logger import crologger
from crodl.tools.scrap import cro_session, get_attributes, get_audio_uuid
from crodl.settings import DOWNLOAD_PATH, PREFERRED_AUDIO_FORMAT, AudioFormat


class AudioWork:
    """
    Processes the audiowork at given URL or by its UUID.

    Args:
        url (str) -- URL of a website with audiowork "hidden" in e.g.:
            https://www.mujrozhlas.cz/hra-na-sobotu/ondrej-neff-velka-solarni-v-nervy-drasajicim-zavode-kosmickych-lodi-jde-doslova-o

        uuid (str) -- Unique ID of the audiowork.
    """

    def __init__(self, **kwargs) -> None:
        _url = kwargs.pop("url", None)
        _uuid = kwargs.pop("uuid", None)
        _title = kwargs.pop("title", None)
        _audiowork_dir = kwargs.pop("audiowork_dir", None)
        _audiowork_root = kwargs.pop("audiowork_root", None)
        _since = kwargs.pop("since", "")

        if _url and _uuid:
            _err_msg = "Audio cannot be defined by both url and uuid!"
            crologger.error(_err_msg)
            raise ValueError(_err_msg)

        if not _url and not _uuid:
            _err_msg = "Audio must be defined by either url or uuid!"
            crologger.error(_err_msg)
            raise ValueError(_err_msg)

        self.url = _url
        self.uuid = _uuid if _uuid else get_audio_uuid(self.url, cro_session)
        self._attrs = get_attributes(self.uuid, cro_session)

        if not _title:
            self.title = self._attrs.get("title", "Unknown")
        else:
            self.title = _title

        if _audiowork_dir:
            self.audiowork_dir = _audiowork_dir
        else:
            self.audiowork_dir = DOWNLOAD_PATH / process_audiowork_title(self.title)

        if _audiowork_root:
            self.audiowork_root = _audiowork_root
        else:
            self.audiowork_root = self.audiowork_dir

        self.series: bool = kwargs.pop("series", False)
        self.show: bool = kwargs.pop("show", False)

        self.since = _since if _since else self._attrs["since"]

    @property
    def audio_links(self) -> list[dict] | None:
        audio_links = self._attrs.get("audioLinks")

        if audio_links:
            return audio_links

        print(f"❌ {self.title}")
        err = "Link not found. This episode is not available."
        crologger.error(self.title)
        crologger.error(err)

        not_yet = not_available_yet(self)
        print(not_yet)

        return None

    @property
    def audio_formats(self) -> list[str] | None:
        audio_variants = []
        if self.audio_links and isinstance(self.audio_links, list):
            for link in self.audio_links:
                if link.get("variant"):
                    audio_variants.append(link.get("variant"))

        if audio_variants and isinstance(audio_variants, list):
            return audio_variants

        return None

    @property
    def audio_formats_urls(self) -> dict[str | None, str | None]:
        """URLs of various audio formats"""
        if self.audio_links and isinstance(self.audio_links, list):
            return {link.get("variant"): link.get("url") for link in self.audio_links}
        return {}

    def info(self):
        """Some basic info on the file being downloaded."""
        attrs = self._attrs
        _description = attrs.get("description")
        _audio_links = attrs["audioLinks"]

        print(f"\n[bold yellow]{self.title}[/bold yellow]")
        for alink in _audio_links:
            bitrate = alink["bitrate"]
            duration = alink["duration"]
            size = alink.get("sizeInBytes", "Stream")
            variant = alink["variant"]

            print(
                f"+++ {HMS(duration)} +++ {file_size(size)} +++ {bitrate} kbps +++ {variant}"
            )

        print(f"\n[blue]{remove_html_tags(_description)}[/blue]\n")

    def already_exists(self) -> bool:  # pragma: no cover
        """Checks whether the audiowork already exists in the download directory."""
        # Get a list of all files in the audiowork directory
        try:
            files_in_directory = os.listdir(self.audiowork_dir)
        except FileNotFoundError:
            files_in_directory = []

        if files_in_directory:
            for file in files_in_directory:
                # Check if the file name matches the provided title (without extension)
                if process_audiowork_title(self.title) in os.path.splitext(file)[0]:
                    return True
        return False

    async def _download_dash(self) -> None:  # pragma: no cover
        """Download audio file from DASH stream
        (m4s segments -> m4a using its manifest.mpd file.)"""
        mpd_url = self.audio_formats_urls.get("dash")

        if not mpd_url:
            raise ValueError("DASH Manifest URL not found.")

        manifest = DASH(
            url=mpd_url, audio_title=self.title, audiowork_dir=self.audiowork_dir
        )

        await manifest.download()

    async def _download_hls(self) -> None:  # pragma: no cover
        """Download audio file from HLS stream
        (aac chunks -> aac file, using its chunklist)."""
        hls_url = self.audio_formats_urls.get("hls")

        if not hls_url:
            raise ValueError("HLS chunklist.txt URL not found.")

        chunklist = HLS(
            url=hls_url, audio_title=self.title, audiowork_dir=self.audiowork_dir
        )
        await chunklist.download()

    def _download_mp3(self):  # pragma: no cover
        """Download mp3 file."""
        mp3_url = self.audio_formats_urls.get("mp3")

        if not mp3_url:
            raise ValueError("MP3 file URL not found.")

        mp3 = MP3(
            url=mp3_url,
            audiowork_dir=self.audiowork_dir,
            audio_title=self.title,
            segments=False,
        )
        mp3.download()

    async def download(
        self, audio_format: Optional[AudioFormat] = PREFERRED_AUDIO_FORMAT
    ) -> None:  # pragma: no cover
        """Method to download the audio file. The download method is picked according
        to the available and preferred audio format.

        If the file already exists, the method will skip it. (Useful for series.)
        """
        if not self.audio_formats:
            return

        if audio_format and audio_format.value not in self.audio_formats:
            # Search for the first preferred available format from class AudioFormat
            crologger.info("The format %s is not available.", audio_format.value)
            audio_format = get_preferred_audio_format(self.audio_formats)
            if audio_format:
                crologger.info("Going to use %s instead.", audio_format.value)

        if not self.already_exists():
            if not self.series and not self.show:
                # Create a download dir for the audiowork when not part of series.
                create_dir_if_does_not_exist(self.audiowork_dir)

            match audio_format:
                case AudioFormat.DASH:
                    await self._download_dash()
                case AudioFormat.HLS:
                    await self._download_hls()
                case AudioFormat.MP3:
                    self._download_mp3()
                case None:
                    err_msg = f"The episdode {self.title} is not available."
                    crologger.error(err_msg)

            crologger.info("Done.")

        else:
            print(f"{self.title} již existuje.")
            crologger.info("%s already exists.", self.title)



from abc import ABC, abstractmethod
from dataclasses import dataclass, field


@dataclass
class Content(ABC):
    url: str
    title: str = field(init=False)

    @abstractmethod
    def already_exists(self) -> bool:
        pass

    @abstractmethod
    async def download(self) -> None:
        pass



from pathlib import Path
from enum import Enum

HOME_DIR = Path.home()
DOWNLOAD_DIR = "Z Rozhlasu"
SEGMENTS_SUBDIR = ".chunks"
DOWNLOAD_PATH = HOME_DIR / DOWNLOAD_DIR
LOG_PATH = DOWNLOAD_PATH / "logs"
SUPPORTED_DOMAINS = ("www.mujrozhlas.cz", "mujrozhlas.cz")
SUPPORTED_AUDIO_FORMATS = ("aac", "m4a")
AUDIO_FORMATS = SUPPORTED_AUDIO_FORMATS + ("mp3",)


class AudioFormat(Enum):
    """The order determines the order of the audio downloads, if available."""

    MP3 = "mp3"
    HLS = "hls"
    DASH = "dash"


PREFERRED_AUDIO_FORMAT = AudioFormat.MP3

API_SERVER = "https://api.mujrozhlas.cz/"



import os

from dataclasses import dataclass
from pathlib import Path

from yaspin import yaspin
from crodl.tools.logger import crologger

from crodl.streams.utils import get_m4a_url, process_audiowork_title
from crodl.streams.audioparts import AudioParts
from crodl.streams.download import download_parts


@dataclass
class HLS(AudioParts):
    """
    Processes a HLS stream using its chunklist, downloaded from its url. Creates
    a local chunklist.txt containing a complete list of all aac chunks which
    will be merged into one aac audio file.
    """

    @property
    def chunklist_path(self) -> Path:
        if not self.segments_path:
            raise ValueError("self.segments_path is None!")
        return self.segments_path / "chunklist.m3u8"

    def _get_chunklist_m3u8(self) -> None:
        from crodl.tools.scrap import cro_session

        mp4_url = get_m4a_url(self.url)
        chunklist_url = mp4_url + "/chunklist.m3u8"
        chunklist = cro_session.get(chunklist_url, timeout=5)
        chunklist.raise_for_status()

        crologger.info("mp4 URL: %s", mp4_url)
        crologger.info("CHUNKLIST URL: %s", chunklist_url)

        with open(self.chunklist_path, "w", encoding="utf-8") as f:
            f.write(chunklist.text)

    def _get_chunk_names(self) -> list[str]:
        with open(self.chunklist_path, encoding="utf-8") as file:
            chunks = [line.rstrip() for line in file if line.startswith("media_")]

        return chunks

    def _delete_chunklist_txt(self) -> None:
        if not self.segments_path:
            raise ValueError("self.segments_path is None!")
        crologger.info("Deleting chunklist.txt.")
        os.remove(self.segments_path / "chunklist.txt")

    def _create_list_txt(self) -> None:
        if not self.segments_path:
            raise ValueError("self.segments_path is None!")

        self._get_chunklist_m3u8()
        chunks = self._get_chunk_names()
        crologger.info("Getting file names...")

        if not chunks:
            raise ValueError("Chunklist is empty!")

        crologger.info("Renaming...")

        if not isinstance(self.segments_path, Path):
            self.segments_path = Path(self.segments_path)
        list_path = Path(self.segments_path / "list.txt")

        with list_path.open("w", encoding="utf-8") as list_txt:
            for line in chunks:
                list_txt.write(line + "\n")
        crologger.info("The list.txt file with new titles was created successfully.")

    def chunks_urls(self) -> list[str]:
        mp4_url = get_m4a_url(self.url)
        chunks = self._get_chunk_names()

        return [mp4_url + "/" + chunk for chunk in chunks]

    def _merge_chunks(self, audio_format: str) -> None:
        if not self.audiowork_dir:
            raise ValueError("self.audiowork_dir is None!")
        if not self.segments_path:
            raise ValueError("self.segments_path is None!")
        if not isinstance(self.audiowork_dir, Path):
            self.audiowork_dir = Path(self.audiowork_dir)

        title_and_extension = (
            f"{process_audiowork_title(self.audio_title)}.{audio_format}"
        )
        output_file = self.audiowork_dir / title_and_extension
        segment_files = [
            self.segments_path / f"media_{i}.{audio_format}"
            for i in range(len(self.chunks_urls()))
        ]

        crologger.info("Merging segments...")
        with open(output_file, "wb") as output:
            for segment in segment_files:
                # crologger.info("Merging segments %s: ", segment)
                with open(segment, "rb") as f:
                    output.write(f.read())

    async def download(self) -> None:
        if not self.segments_path:
            raise ValueError("self.segments_path is not set!")

        with yaspin(text=f"Zpracovávám {self.audio_title}", color="yellow") as _:
            self._create_list_txt()

        with yaspin(text=f"Stahuji {self.audio_title}", color="yellow") as _:
            await download_parts(self.chunks_urls(), self.segments_path)

        with yaspin(text=f"Ukládám {self.audio_title}", color="yellow") as spinner:
            self._merge_chunks("aac")
            self._purge_chunks_dir()
            spinner.ok("✔️")
